@@ -22,8 +22,12 @@ import weakref
 import functools
 import operator
 import warnings
+import inspect
+import random
+import textwrap
+import types
 
-__version__ = "0.6.0"  # automatically patched by setup.py when packaging
+__version__ = "0.8.0"  # automatically patched by setup.py when packaging
 
 # from six.py's strategy
 INTEGER_TYPES = None
@@ -109,6 +113,7 @@ CALL = "call"
 IMPORT = "import"
 DEL = "del"
 EVAL = "eval"
+EXEC = "exec"
 EXPR = "expr"
 RESULT = "result"
 ERROR = "error"
@@ -695,6 +700,7 @@ class BridgeConn(object):
         # if the bridge has requested a local_call_hook/local_eval_hook, record that
         self.local_call_hook = bridge.local_call_hook
         self.local_eval_hook = bridge.local_eval_hook
+        self.local_exec_hook = bridge.local_exec_hook
 
         if record_stats:
             self.stats = Stats()
@@ -764,7 +770,9 @@ class BridgeConn(object):
             serialized_dict = {TYPE: INT, VALUE: str(data)}
         elif isinstance(data, float):
             serialized_dict = {TYPE: FLOAT, VALUE: str(data)}
-        elif isinstance(data, STRING_TYPES):  # all strings are coerced to unicode
+        elif isinstance(
+            data, STRING_TYPES
+        ):  # all strings are coerced to unicode when serialized
             serialized_dict = {
                 TYPE: STR,
                 VALUE: base64.b64encode(data.encode("utf-8")).decode("utf-8"),
@@ -842,7 +850,18 @@ class BridgeConn(object):
         elif serial_dict[TYPE] == BOOL:
             return serial_dict[VALUE] == "True"
         elif serial_dict[TYPE] == STR:
-            return base64.b64decode(serial_dict[VALUE]).decode("utf-8")
+            result = base64.b64decode(serial_dict[VALUE]).decode("utf-8")
+            # if we're in python 2, result is now a unicode string.
+            if sys.version_info[0] == 2:
+                try:
+                    # We'll try and force it down to a plain string, because there are plenty of cases where plain strings
+                    # are expected instead of unicode (e.g., type/module names). If that fails, we'll keep it as unicode.
+                    result = str(result)
+                except UnicodeEncodeError:
+                    # couldn't make it ascii, keep as unicode
+                    pass
+            return result
+
         elif serial_dict[TYPE] == BYTES:
             return base64.b64decode(serial_dict[VALUE])
         elif serial_dict[TYPE] == LIST:
@@ -1302,7 +1321,7 @@ class BridgeConn(object):
                         e, traceback.format_exc()
                     )
                 )
-                self.logger.debug("Falling back:\n\local_eval {}".format(args_dict))
+                self.logger.debug("Falling back:\nlocal_eval {}".format(args_dict))
 
         try:
             """ the import __main__ trick allows accessing all the variables that the bridge imports, 
@@ -1325,6 +1344,118 @@ class BridgeConn(object):
 
         return result
 
+    @stats_hit
+    def remote_exec(self, exec_string, timeout_override=None, **kwargs):
+        self.logger.debug("remote_exec({}, {})".format(exec_string, kwargs))
+
+        command_dict = {
+            CMD: EXEC,
+            ARGS: self.serialize_to_dict({EXPR: exec_string, KWARGS: kwargs}),
+        }
+        # Remote exec commands might take a while, so override the timeout value, factor 100 is arbitrary unless an override specified by caller
+        if timeout_override is None:
+            timeout_override = self.response_timeout * 100
+        result = self.send_cmd(command_dict, timeout_override=timeout_override)
+
+        return self.deserialize_from_dict(result)
+
+    @stats_hit
+    def local_exec(self, args_dict):
+        args = self.deserialize_from_dict(args_dict)
+
+        result = None
+
+        if self.logger.getEffectiveLevel() <= logging.DEBUG:
+            try:
+                # we want to get log the deserialized values, because they're useful.
+                # but this also means a bad repr can break things. So we get ready to
+                # catch that and fallback to undeserialized values
+                self.logger.debug("local_exec({},{})".format(args[EXPR], args[KWARGS]))
+            except Exception as e:
+                self.logger.debug(
+                    "Failed to log deserialized arguments: {}\n{}".format(
+                        e, traceback.format_exc()
+                    )
+                )
+                self.logger.debug("Falling back:\nlocal_exec {}".format(args_dict))
+
+        try:
+            """ the import __main__ trick allows accessing all the variables that the bridge imports, 
+            so execs will run within the global context of what started the bridge, and the arguments 
+            supplied as kwargs will override that """
+            exec_expr = args[EXPR]
+            exec_globals = importlib.import_module("__main__").__dict__
+            # unlike remote_eval, we add the kwargs to the globals, because the most common use of remote_exec is to define a function/class, and locals aren't accessible in those definitions
+            exec_globals.update(args[KWARGS])
+            # do the exec, or defer to the hook if we've registered one
+            if self.local_exec_hook is None:
+                exec(exec_expr, exec_globals)
+            else:
+                self.local_exec_hook(self, exec_expr, exec_globals)
+            self.logger.debug("local_exec: Finished executing")
+        except Exception as e:
+            result = e
+            traceback.print_exc()
+
+        return result
+
+    def remoteify(self, module_class_or_function, **kwargs):
+        """ Push a module, class or function definition into the remote python interpreter, and return a handle to it.
+
+            Notes: 
+                * requires that the class or function code is able to be understood by the remote interpreter (e.g., if it's running python2, the source must be python2 compatible)
+                * If remoteify-ing a class, the class can't be defined in a REPL (a limitation of inspect.getsource). You need to define it in a file somewhere.
+                * If remoteify-ing a module, it can't do relative imports - they require a package structure which won't exist
+                * If remoteify-ing a module, you only get the handle back - it's not installed into the remote or local sys.modules, you need to do that yourself.
+                * You can't remoteify a decorated function/class - it'll only get the source for the decorator wrapper, not the original.
+        """
+        source_string = inspect.getsource(module_class_or_function)
+        name = module_class_or_function.__name__
+
+        # random name that'll appear in the __main__ globals to retrieve the remote definition.
+        # Used to avoid colliding with other uses of the name that might be there, or other clients
+        temp_name = "_bridge_remoteify_temp_result" + "".join(
+            [random.choice("0123456789ABCDEF") for _ in range(0, 8)]
+        )
+
+        if isinstance(module_class_or_function, types.ModuleType):
+            """ Modules need a bit of extra love and care. """
+            # We'll use the temp_name to store the source of the module (makes it easier than patching it into the format string below and escaping everything),
+            # and pass it as a global to the exec
+            kwargs[temp_name] = source_string
+
+            # We create a new module context to execute the module code in, then run a second exec from
+            # the first exec inside the new module's __dict__, so imports are set correctly as globals of the module (not globals of the exec)
+            # Note that we need to force the module name to be a string - python2 doesn't support unicode module names
+            source_string = "import types\nnew_mod = types.ModuleType(str('{name}'))\nexec({temp_name}, new_mod.__dict__)\n".format(
+                name=name, temp_name=temp_name
+            )
+            # update name to capture the new module object we've created
+            name = "new_mod"
+
+        elif (
+            source_string[0] in " \t"
+        ):  # modules won't be indented, only a class/function issue
+            # source is indented to some level, so dedent it to avoid an indentation error
+            source_string = textwrap.dedent(source_string)
+
+        retrieval_string = "\nglobals()['{temp_name}'] = {name}".format(
+            temp_name=temp_name, name=name
+        )
+
+        # run the exec
+        self.remote_exec(source_string + retrieval_string, **kwargs)
+
+        # retrieve from __main__ with remote_eval
+        result = self.remote_eval(temp_name)
+
+        # nuke the temp name - the remote handle will keep the module/class/function around
+        self.remote_exec(
+            "global {temp_name}\ndel {temp_name}".format(temp_name=temp_name)
+        )
+
+        return result
+
     @stats_hit
     def remote_shutdown(self):
         self.logger.debug("remote_shutdown")
@@ -1378,6 +1509,8 @@ class BridgeConn(object):
                 result = self.local_isinstance(command_dict[ARGS])
             elif command_dict[CMD] == EVAL:
                 result = self.local_eval(command_dict[ARGS])
+            elif command_dict[CMD] == EXEC:
+                result = self.local_exec(command_dict[ARGS])
             elif command_dict[CMD] == SHUTDOWN:
                 result = self.local_shutdown()
 
@@ -1445,7 +1578,7 @@ class BridgeConn(object):
 
         local_type = type(
             str("_bridged_" + type_name), bases, class_dict
-        )  # str to force it to non-unicode in py2
+        )  # str to force it to non-unicode in py2 (is unicode thanks to unicode_literals)
         self.cached_bridge_types[type_name] = local_type
 
         return local_type
@@ -1483,6 +1616,7 @@ class BridgeServer(
     is_serving = False
     local_call_hook = None
     local_eval_hook = None
+    local_exec_hook = None
 
     def __init__(
         self,
@@ -1492,6 +1626,7 @@ class BridgeServer(
         response_timeout=DEFAULT_RESPONSE_TIMEOUT,
         local_call_hook=None,
         local_eval_hook=None,
+        local_exec_hook=None,
     ):
         """ Set up the bridge.
 
@@ -1528,6 +1663,7 @@ class BridgeServer(
         # hook local_call/local_eval to allow inspection/modification of calls/evals (e.g., forcing them onto a particular thread)
         self.local_call_hook = local_call_hook
         self.local_eval_hook = local_eval_hook
+        self.local_exec_hook = local_exec_hook
 
     def get_server_info(self):
         """ return where the server is serving on """
@@ -1562,6 +1698,7 @@ class BridgeClient(object):
 
     local_call_hook = None
     local_eval_hook = None
+    local_exec_hook = None
     _bridge = None
 
     def __init__(
@@ -1635,6 +1772,29 @@ class BridgeClient(object):
             eval_string, timeout_override=timeout_override, **kwargs
         )
 
+    def remote_exec(self, exec_string, timeout_override=None, **kwargs):
+        """ Takes python script as a string and executes it entirely on the server.
+
+            To provide arguments into the exec context, supply them as keyword arguments with names matching the names used in the exec string (e.g., remote_exec("print(x)", x="helloworld")).
+
+            Note: the python script must be able to be understood by the remote interpreter (e.g., if it's running python2, the script must be python2 compatible)
+        """
+        return self.client.remote_exec(
+            exec_string, timeout_override=timeout_override, **kwargs
+        )
+
+    def remoteify(self, module_class_or_function, **kwargs):
+        """ Push a module, class or function definition into the remote python interpreter, and return a handle to it.
+
+            Notes: 
+                * requires that the class or function code is able to be understood by the remote interpreter (e.g., if it's running python2, the source must be python2 compatible)
+                * If remoteify-ing a class, the class can't be defined in a REPL (a limitation of inspect.getsource). You need to define it in a file somewhere.
+                * If remoteify-ing a module, it can't do relative imports - they require a package structure which won't exist
+                * If remoteify-ing a module, you only get the handle back - it's not installed into the remote or local sys.modules, you need to do that yourself.
+                * You can't remoteify a decorated function/class - it'll only get the source for the decorator wrapper, not the original.
+        """
+        return self.client.remoteify(module_class_or_function, **kwargs)
+
     def remote_shutdown(self):
         return self.client.remote_shutdown()
 
